
#  --------------------------------------------------------------------------------
#  Izok: A modeling system similar to GAMS for pyomo.
#  Copyright 2024,  Andrey Sergeevich Storozhenko, All rights reserved.
#  This software is distributed under the 3-clause BSD License.
#  --------------------------------------------------------------------------------

# parser error_id const
PARSER_EXPECT_EXCEPTION = 1
PARSER_NOT_UNARY_OPERATOR = 2
PARSER_MATCH_RULE_EXCEPTION = 3
PARSER_BAD_CHARS = 4
PARSER_OBJ_SENSE_ERROR = 5
PARSER_DOMAIN_DEFINITION_ERROR = 6

# generation error
GEN_TYPE_CONVERT_ERROR = 7
GEN_FUNCTION_CALL_ERROR = 8
GEN_PROPERTY_NAME_ERROR = 9
GEN_OPERATOR_ERROR = 10
GEN_ESP_NONE_ERROR = 11
GEN_INDEXING_ERROR = 12



#  --------------------------------------------------------------------------------
#  Izok: A modeling system similar to GAMS for pyomo.
#  Copyright 2024,  Andrey Sergeevich Storozhenko, All rights reserved.
#  This software is distributed under the 3-clause BSD License.
#  --------------------------------------------------------------------------------

from typing import Union, List
from izok.ast_est import AST, EST
from izok.ast import OperandAST, IdentListAST, ExprAST
from izok.baseparser import ParserRule, BaseParser
from izok.token import *
from .domain import Domain
from .model import init_constraint, skip_constraint


class EquationAST(AST):
    def __init__(self,
                 pos: Position, name: str,
                 domain: Union[AST, None],
                 expr: AST):
        super(EquationAST, self).__init__(pos)
        self.name = name
        self.domain = domain
        self.expr = expr

    def __str__(self):
        return f"{self.name} {'' if self.domain is None else self.domain} .. {self.expr} ;"

    def to_est(self) -> Union[EST, None]:
        if isinstance(self.domain, OperandAST):
            domain = self.domain.expr
            cond = self.domain.cond.to_est()
        else:
            domain = self.domain
            cond = None

        if isinstance(domain, IdentListAST):
            return EquationEST(self, self.name, domain.idents, cond, self.expr.to_est())
        else:
            return EquationEST(self, self.name, domain, cond, self.expr.to_est())


class EquationEST(EST):
    def __init__(self, ast: AST, ident: str, domain: Union[List[str], None], cond: Union[EST, None], expr: EST):
        super(EquationEST, self).__init__(ast)
        self.ident = ident
        self.domain = domain
        self.cond = cond
        self.expr = expr

    def __call__(self, model, *args):
        domain = Domain()
        if self.domain is None:
            return self.expr(model, domain)
        else:
            # init domain from args
            if self.cond is not None:
                if self.cond(model, domain):
                    return self.expr(model, domain)
                else:
                    return skip_constraint()
            else:
                return self.expr(model, domain)

    def init(self, model):
        init_constraint(model, self.ast, self.domain, self.ident, self)


class EquationRule(ParserRule):
    def __init__(self):
        super(EquationRule, self).__init__("equation")

    def parse(self, parser: BaseParser) -> AST:
        # name domain.. expr eq_type expr ;

        name = parser.scaner.token.value
        pos = parser.scaner.token.pos
        parser.scan()

        domain = parser.domain()

        parser.expect(TOKEN_POINTPOINT)
        left = parser.expr()
        eq_type = parser.expect(TOKEN_EQNTYPE)
        right = parser.expr()

        parser.expect(TOKEN_SEMICOLON)
        return EquationAST(pos, name, domain, ExprAST(eq_type.pos, eq_type.value, left, right))

    def skip_block(self, parser: BaseParser) -> None:
        parser.skip_to_semicolon_and_skip()



#  --------------------------------------------------------------------------------
#  Izok: A modeling system similar to GAMS for pyomo.
#  Copyright 2024,  Andrey Sergeevich Storozhenko, All rights reserved.
#  This software is distributed under the 3-clause BSD License.
#  --------------------------------------------------------------------------------

from typing import Union, Callable, List
from pyomo.environ import *
from .domain import Domain
from izok.ast_est import AST
from izok.error import error
from izok.errorid import *


def get_property_by_name(model: Model, domain: Union[Domain, None], ast: AST,  ident: str):
    if domain is not None and ident in domain:
        return domain.get_set_value(ident)
    else:
        try:
            return getattr(model, ident)
        except AttributeError:
            error(ast.pos,
                  GEN_PROPERTY_NAME_ERROR,
                  "Error in model property name",
                  f"Error in model property name: {ident}")


def is_set(obj):
    return isinstance(obj, Set)


def get_set(model: Model, ast: AST, property_name: str) -> Union[Set, None]:
    prp_set = get_property_by_name(model, None, ast, property_name)

    if is_set(prp_set):
        return prp_set.value
    else:
        return None


def check_index_or_call(obj) -> bool:
    if isinstance(obj, Var) or isinstance(obj, Param):
        return True
    elif isinstance(obj, Callable):
        return False
    else:
        Exception("Panic. Object not indexing and can be called.")


def skip_constraint():
    return Constraint.Skip


def init_constraint(model: Model, ast: AST, domain: [List[str], None], ident: str, eq: Callable):
    if domain is None:
        setattr(model, ident, Constraint(rule=eq))
    else:
        domain_lst = []
        for set_name in domain:
            domain_lst.append(get_set(model, ast, set_name))

        setattr(model, ident, Constraint(*domain_lst, rule=eq))


def init_objective(model: Model, ast: AST, ident: str, sense: str, obj: Callable):
    if sense == 'MAXIMIZE':
        setattr(model, ident, Objective(rule=obj, sense=maximize))
    else:
        setattr(model, ident, Objective(rule=obj, sense=minimize))



#  --------------------------------------------------------------------------------
#  Izok: A modeling system similar to GAMS for pyomo.
#  Copyright 2024,  Andrey Sergeevich Storozhenko, All rights reserved.
#  This software is distributed under the 3-clause BSD License.
#  --------------------------------------------------------------------------------

from .position import Position
from .est import *


class IdentAST(AST):
    def __init__(self,  pos: Position, ident: str):
        super(IdentAST, self).__init__(pos)
        self.ident = ident

    def __str__(self):
        return self.ident

    def to_est(self) -> Union[EST, None]:
        return IdentEST(self, self.ident)


class LabelAST(AST):
    def __init__(self,  pos: Position, value: str):
        super(LabelAST, self).__init__(pos)
        self.value = value

    def __str__(self):
        return f"'{self.value}'"

    def to_est(self) -> Union[EST, None]:
        return LabelEST(self, self.value)


class NumberAST(AST):
    def __init__(self,  pos: Position, value: str):
        super(NumberAST, self).__init__(pos)
        self.value = value

    def __str__(self):
        return self.value

    def to_est(self) -> Union[EST, None]:
        return NumberEST(self, self.value)


class ExprAST(AST):
    def __init__(self, pos: Position, value: str, left_op: Union[AST, None], right_op: AST):
        super(ExprAST, self).__init__(pos)
        self.value = value
        self.left = left_op
        self.right = right_op

    def __str__(self):
        if self.left is None:
            return f"[{self.value} {str(self.right)}]"
        else:
            return f"[{str(self.left)} {self.value} {str(self.right)}]"

    def to_est(self) -> Union[EST, None]:
        if self.left is None:
            return ExprEST(self, self.value, None, self.right.to_est())
        else:
            return ExprEST(self, self.value, self.left.to_est(), self.right.to_est())


class OperandAST(AST):
    def __init__(self, expr: AST, cond: Union[AST, None]):
        super(OperandAST, self).__init__(expr.pos)
        self.expr = expr
        self.cond = cond

    def __str__(self):
        return f"{self.expr}{'' if self.cond is None else '$'+str(self.cond)}"

    def to_est(self) -> Union[EST, None]:
        if self.cond is None:
            return OperandEST(self, self.expr.to_est(), None)
        else:
            return OperandEST(self, self.expr.to_est(), self.cond.to_est())


class SubExprAST(AST):
    def __init__(self, pos: Position, expr: Union[AST, None]):
        super(SubExprAST, self).__init__(pos)
        self.expr = expr

    def __str__(self):
        return f"({'' if self.expr is None else self.expr})"

    def to_est(self) -> Union[EST, None]:
        if self.expr is None:
            return None
        else:
            return self.expr.to_est()


class IdentListAST(AST):
    def __init__(self, pos: Position, idents: List[str]):
        super(IdentListAST, self).__init__(pos)
        self.idents = idents

    def __str__(self):
        if len(self.idents) == 1:
            return f"{self.idents[0]}"
        else:
            return f"({','.join(str(x) for x in self.idents)})"

    def to_est(self) -> Union[EST, None]:
        return SetListEST(self, self.idents)


class ExprListAST(AST):
    def __init__(self, pos: Position, expr_list: List[AST]):
        super(ExprListAST, self).__init__(pos)
        self.expr_list = expr_list

    def __str__(self):
        if len(self.expr_list) == 1:
            return f"{self.expr_list[0]}"
        else:
            return f"{','.join(str(x) for x in self.expr_list)}"

    def to_est(self) -> Union[EST, None]:
        expr_list = []
        for expr in self.expr_list:
            expr_est = expr.to_est()
            if expr_est is None:
                error(expr.pos,
                      GEN_ESP_NONE_ERROR,
                      "Convert to a EST error",
                      "When converted to a EST, the result is None"
                      )
            expr_list.append(expr_est)

        return ExprListEST(self, expr_list)


class FunctionAST(AST):
    def __init__(self, pos: Position, ident: str, args: Union[IdentListAST, ExprListAST]):
        super(FunctionAST, self).__init__(pos)
        self.ident = ident
        self.args = args

    def __str__(self):
        return f"{self.ident}({self.args})"

    def to_est(self) -> Union[EST, None]:
        return FunctionEST(self, self.ident, self.args.to_est())



#  --------------------------------------------------------------------------------
#  Izok: A modeling system similar to GAMS for pyomo.
#  Copyright 2024,  Andrey Sergeevich Storozhenko, All rights reserved.
#  This software is distributed under the 3-clause BSD License.
#  --------------------------------------------------------------------------------

class Box:
    def __init__(self, name: str, owner=None):
        self._owner = owner
        self._name = name
        self.value = None

    @property
    def name(self):
        return self._name

    @property
    def owner(self):
        return self._owner


class Domain:
    def __init__(self):
        self._domain = dict()

    def push_set(self, name: str, owner):
        self._domain[name] = [Box(name, owner)]

    def pop_set(self, name: str):
        self._domain[name].pop(-1)

    def get_set_by_name(self, name: str) -> Box:
        return self._domain[name][-1]

    def get_set_value(self, name: str):
        return self._domain[name][-1].value

    def set_set_value(self, name: str, value):
        self._domain[name][-1].value = value

    def __contains__(self, item: str):
        return item in self._domain



#  --------------------------------------------------------------------------------
#  Izok: A modeling system similar to GAMS for pyomo.
#  Copyright 2024,  Andrey Sergeevich Storozhenko, All rights reserved.
#  This software is distributed under the 3-clause BSD License.
#  --------------------------------------------------------------------------------

from .ast import *
from .scaner import Scaner, un_op_names, operator_priority
from .token import *
from .error import IzokException, error, set_base_position
from .errorid import *


class ParserRule:
    def __init__(self, ident: str):
        self.ident = ident

    def act_parse(self, parser) -> AST:
        set_base_position(parser.scaner.token.pos)
        ast = self.parse(parser)
        set_base_position(None)
        return ast

    def parse(self, parser) -> AST:
        raise Exception("Must be overriden")

    def skip_block(self, parser) -> None:
        raise Exception("Must be overriden")


class FunctionRule:
    def __init__(self, ident: str):
        self.ident = ident

    def parse(self, parser, ident: Token) -> AST:
        raise Exception("Must be overriden")


class BaseParser:
    def __init__(self, text: str,
                 rules: List[ParserRule], default_rule: ParserRule = None,
                 function_rules: List[FunctionRule] = None):
        self.scaner = Scaner(text)
        self._rules = dict()
        for rule in rules:
            self._rules[rule.ident.upper()] = rule
        self._default_rule = default_rule
        self._function_rules = dict()
        if function_rules is not None:
            for function_rule in function_rules:
                self._function_rules[function_rule.ident.upper()] = function_rule

    def error(self, error_id: int, head_msg: str, tile_msg: str, base_token: Union[Token, None] = None) -> None:
        if base_token is None:
            error(self.scaner.token.pos, error_id, head_msg, tile_msg)
        else:
            error(base_token.pos, error_id, head_msg, tile_msg)

    def scan(self) -> None:

        def step():
            self.scaner.scan()
            bad_chars = ""
            token_begin = None
            while self.scaner.token.type == "":
                if token_begin is None:
                    token_begin = self.scaner.token
                bad_chars += self.scaner.token.value
                self.scaner.scan()
            if bad_chars != "":
                self.error(PARSER_EXPECT_EXCEPTION,
                           "Bad chars",
                           f"Bad chars {bad_chars}", token_begin)

        step()
        while self.case(TOKEN_WHITE_SPACE, TOKEN_NEW_LINE):
            step()

    def case(self, *tokens: str) -> bool:
        return any(self.scaner.token.type == t for t in tokens)

    def expect(self, *tokens: str) -> Token:
        if not any(self.scaner.token.type == token for token in tokens):
            self.error(PARSER_EXPECT_EXCEPTION,
                       "Unexpected token",
                       f"Expect {','.join(token.upper() for token in tokens)} " +
                       f"but got {self.scaner.token.type.upper()}")
        ret = self.scaner.token
        self.scan()
        return ret

    def skip_to_semicolon_and_skip(self) -> None:
        while not self.case(TOKEN_SEMICOLON, TOKEN_END):
            self.scan()
        if self.case(TOKEN_SEMICOLON):
            self.scan()

    def parse(self) -> Union[AST, None]:
        if self.scaner.token is None:
            self.scan()

        if self.case(TOKEN_END):
            return None

        if self.case(TOKEN_IDENT):
            rule = self._rules.get(self.scaner.token.value.upper(), None)
            if rule is None:
                if self._default_rule is None:
                    self.error(PARSER_MATCH_RULE_EXCEPTION,
                               "Rules key error.",
                               f"The identifier name ({self.scaner.token.value}) does not match any rules.")
                else:
                    try:
                        return self._default_rule.act_parse(self)
                    except IzokException as err:
                        self._default_rule.skip_block(self)
                        print(err)
            else:
                try:
                    return rule.act_parse(self)
                except IzokException as err:
                    rule.skip_block(self)
                    print(err)
        else:
            self.expect(TOKEN_IDENT)
        return None

    def expr(self) -> AST:
        stack = []
        line = []

        if self.case(TOKEN_OPERATOR):
            if not (self.scaner.token.value in un_op_names):
                self.error(PARSER_NOT_UNARY_OPERATOR,
                           "The operator is not a unary operator.",
                           f"The {self.scaner.token.value.upper()} operator is not unary.")
            opr = self.scaner.token
            self.scan()
            line.append(
                ExprAST(opr.pos, opr.value, None, self.operand()))
        else:
            line.append(self.operand())

        while self.case(TOKEN_OPERATOR):
            pr = operator_priority[self.scaner.token.value]
            if len(stack) != 0:
                op1, pr1 = stack[-1]
                if abs(pr) > abs(pr1):
                    stack.pop(-1)
                    if pr1 < 0:
                        line.append(
                            ExprAST(op1.pos, op1.value, None, line.pop(-1)))
                    else:
                        right_op = line.pop(-1)
                        left_op = line.pop(-1)
                        line.append(ExprAST(op1.pos, op1.value, left_op, right_op))
            stack.append((self.scaner.token, pr))
            self.scan()
            line.append(self.operand())

        while len(stack) != 0:
            op, _ = stack.pop(-1)
            right_op = line.pop(-1)
            left_op = line.pop(-1)
            line.append(ExprAST(op.pos, op.value, left_op, right_op))

        return line[0]

    def operand(self) -> AST:
        cond = None
        base = self._prim_operand()
        if self.case(TOKEN_DOLLAR):
            self.scan()
            cond = self.expr()
        if cond is None:
            return base
        else:
            return OperandAST(base, cond)

    def _prim_operand(self) -> AST:
        ret = self.sub_expr()
        if ret is None:
            ret = self.lit()
            if ret is None:
                ret = self.ident()
                if ret is None:
                    if self.scaner.token.value == '\n':
                        end = "END_OF_LINE"
                    elif self.scaner.token.value == '\0':
                        end = "END_OF_STRING"
                    else:
                        end = self.scaner.token.value
                    self.error(PARSER_EXPECT_EXCEPTION,
                               "Unexpected token",
                               f"Expect ( or NUMBER or LABEL or IDENT but got {end}"
                               )
        return ret

    def expr_list(self) -> ExprListAST:
        expr_list = []
        expr = self.expr()
        expr_list.append(expr)
        while self.case(TOKEN_COMMA):
            self.scan()
            expr_list.append(self.expr())
        return ExprListAST(expr_list[0].pos, expr_list)

    def ident(self) -> Union[AST, None]:
        if self.case(TOKEN_IDENT):
            ident = self.scaner.token
            self.scan()
            if self.case(TOKEN_LPAREN):
                self.scan()

                function_rule = self._function_rules.get(ident.value.upper(), None)

                if function_rule is None:
                    expr_list = self.expr_list()
                    self.expect(TOKEN_RPAREN)
                    return FunctionAST(ident.pos, ident.value, expr_list)
                else:
                    ret = function_rule.parse(self, ident)
                    self.expect(TOKEN_RPAREN)
                    return ret
            return IdentAST(ident.pos, ident.value)
        return None

    def lit(self) -> Union[NumberAST, LabelAST, None]:
        if self.case(TOKEN_NUMBER, TOKEN_LABEL):
            if self.scaner.token.type == TOKEN_LABEL:
                ret = LabelAST(self.scaner.token.pos, self.scaner.token.value)
            else:
                ret = NumberAST(self.scaner.token.pos, self.scaner.token.value)
            self.scan()
            return ret
        return None

    def sub_expr(self) -> Union[AST, None]:
        if self.case(TOKEN_LPAREN):
            lbr = self.scaner.token
            self.scan()
            if self.case(TOKEN_RPAREN):
                expr = None
            else:
                expr = self.expr()
            self.expect(TOKEN_RPAREN)
            return SubExprAST(lbr.pos, expr)
        else:
            return None

    def domain(self) -> Union[AST, None]:
        cond = None
        if self.case(TOKEN_IDENT):
            base = IdentAST(self.scaner.token.pos, self.scaner.token.value)
            self.scan()
        elif self.case(TOKEN_LPAREN):
            pos = self.scaner.token.pos
            self.scan()
            if self.case(TOKEN_RPAREN):
                base = IdentListAST(pos, [])
            else:
                idents = [self.expect(TOKEN_IDENT).value]
                while self.case(TOKEN_COMMA):
                    self.scan()
                    idents.append(self.expect(TOKEN_IDENT).value)
                base = IdentListAST(pos, idents)
                self.expect(TOKEN_RPAREN)
        else:
            return None

        if self.case(TOKEN_DOLLAR):
            self.scan()
            cond = self.expr()

        if cond is None:
            return base
        else:
            return OperandAST(base, cond)



#  --------------------------------------------------------------------------------
#  Izok: A modeling system similar to GAMS for pyomo.
#  Copyright 2024,  Andrey Sergeevich Storozhenko, All rights reserved.
#  This software is distributed under the 3-clause BSD License.
#  --------------------------------------------------------------------------------

from .baseparser import BaseParser
from .extension.objective import ObjectiveRule
from .extension.equation import EquationRule
#from izok.extension.sum import SumFunctionRule


class Parser(BaseParser):
    def __init__(self, text: str):
        super(Parser, self).__init__(text,
                                     [
                                        ObjectiveRule()
                                     ], EquationRule(),
                                     [
                                         #SumFunctionRule()
                                     ])



#  --------------------------------------------------------------------------------
#  Izok: A modeling system similar to GAMS for pyomo.
#  Copyright 2024,  Andrey Sergeevich Storozhenko, All rights reserved.
#  This software is distributed under the 3-clause BSD License.
#  --------------------------------------------------------------------------------

class Position:
    def __init__(self, source: str, line: int, pos: int = 0, offset: int = 0):
        self.source = source
        self.line = line
        self.pos = pos
        self.offset = offset



#  --------------------------------------------------------------------------------
#  Izok: A modeling system similar to GAMS for pyomo.
#  Copyright 2024,  Andrey Sergeevich Storozhenko, All rights reserved.
#  This software is distributed under the 3-clause BSD License.
#  --------------------------------------------------------------------------------

from typing import Union
import re
from .token import *
from .basescaner import ScanerRule, BaseScaner

# parser regular expressions
white_space_re = re.compile(r" +")
new_line_re = re.compile(r"(\r\n|\r|\n)")

comment_re = re.compile(r"#(.*)")
ident_re = re.compile(r"(\w+)")
label_re = re.compile(r"'([^'\r\n]*)['\r\n]")
number_re = re.compile(r"(\d+([.]\d*)?([eE][+-]?\d+)?)|([.]\d+([eE][+-]?\d+)?)")

# bind operators to token types
ex_operators = {
    ";": TOKEN_SEMICOLON,
    "(": TOKEN_LPAREN,
    ")": TOKEN_RPAREN,
    ",": TOKEN_COMMA,
    "$": TOKEN_DOLLAR,
    "..": TOKEN_POINTPOINT,
    "=e=": TOKEN_EQNTYPE,
    "=l=": TOKEN_EQNTYPE,
    "=g=": TOKEN_EQNTYPE,
    "=E=": TOKEN_EQNTYPE,
    "=L=": TOKEN_EQNTYPE,
    "=G=": TOKEN_EQNTYPE,
}

un_op_names = ["+", "-", "not"]

operator_priority = {
    "+": 3,
    "-": 3,
    "/": 2,
    "*": 2,
    "**": 1,
    "=": 4,
    "<>": 4,
    ">": 4,
    ">=": 4,
    "<": 4,
    "<=": 4,
    "not": 5,
    "and": 6,
    "or": 7,
    "xor": 7,
}

operator_names = list(set(list(operator_priority) + un_op_names))

all_operators = dict()
for name in operator_names:
    all_operators[name] = TOKEN_OPERATOR

operators = {**all_operators, **ex_operators}

operator_re = re.compile(
    f"({'|'.join(re.escape(x) for x in sorted(operators, key=lambda x: -len(x)))})"
)


class Scaner(BaseScaner):
    def __init__(self, text: str):
        def default_action(pos: Position, token: str, value: str) -> (Token, Position):
            return Token(pos, token, value), Position(pos.source,
                                                      pos.line,
                                                      pos.pos + len(value),
                                                      pos.offset + len(value))

        super(Scaner, self).__init__(text, [
            ScanerRule(white_space_re, TOKEN_WHITE_SPACE, default_action),
            ScanerRule(new_line_re, TOKEN_NEW_LINE,
                       lambda pos, token, value: (Token(pos, token, value),
                                                  Position(self._text_by_line[pos.line],
                                                           pos.line + 1,
                                                           0,
                                                           pos.offset + len(value)))),
            ScanerRule(comment_re, TOKEN_COMMENT,
                       lambda pos, token, value: (Token(pos, token, value[1:]),
                                                  Position(pos.source,
                                                           pos.line,
                                                           pos.pos + len(value),
                                                           pos.offset + len(value)))),
            ScanerRule(label_re, TOKEN_LABEL,
                       lambda pos, token, value: (Token(pos, token, value[1:-1]),
                                                  Position(pos.source,
                                                           pos.line,
                                                           pos.pos + len(value) - 1,
                                                           pos.offset + len(value) - 1))
                       if value[-1] != "'" else (Token(pos, token, value[1:-1]),
                                                 Position(pos.source,
                                                          pos.line,
                                                          pos.pos + len(value),
                                                          pos.offset + len(value)))
                       ),
            ScanerRule(number_re, TOKEN_NUMBER, default_action),
            ScanerRule(operator_re, TOKEN_OPERATOR,
                       lambda pos, token, value: (Token(pos, operators[value], value),
                                                  Position(pos.source,
                                                           pos.line,
                                                           pos.pos + len(value),
                                                           pos.offset + len(value)))),
            ScanerRule(ident_re, TOKEN_IDENT, default_action),
        ])

        self.token = None

    def scan(self) -> Union[Token, None]:
        self.token = super().scan()
        if self.token is None:
            self.token = Token(self.pos, TOKEN_END, '\0')
        return self.token



#  --------------------------------------------------------------------------------
#  Izok: A modeling system similar to GAMS for pyomo.
#  Copyright 2024,  Andrey Sergeevich Storozhenko, All rights reserved.
#  This software is distributed under the 3-clause BSD License.
#  --------------------------------------------------------------------------------

from typing import Union
from izok.ast_est import AST, EST
from izok.baseparser import ParserRule, BaseParser
from izok.token import *
from izok.errorid import *
from .domain import Domain
from .model import init_objective


class ObjectiveAST(AST):
    def __init__(self, pos: Position, name: str, sense: str, expr: AST):
        super(ObjectiveAST, self).__init__(pos)
        self.name = name
        self.sense = sense
        self.expr = expr

    def __str__(self):
        return f"obj {self.sense} {self.name} .. {self.expr} ;"

    def to_est(self) -> Union['EST', None]:
        return ObjectiveEST(self, self.name, self.sense, self.expr.to_est())


class ObjectiveEST(EST):
    def __init__(self, ast: AST, ident: str, sense: str, expr: EST):
        super(ObjectiveEST, self).__init__(ast)
        self.ident = ident
        self.sense = sense
        self.expr = expr

    def __call__(self, model, *args):
        domain = Domain()
        return self.expr(model, domain)

    def init(self, model):
        init_objective(model, self.ast, self.ident, self.sense, self)


class ObjectiveRule(ParserRule):
    def __init__(self):
        super(ObjectiveRule, self).__init__("obj")

    def parse(self, parser: BaseParser) -> AST:
        # obj maximize | minimize name .. expr ;
        pos = parser.scaner.token.pos
        parser.scan()
        sense = parser.expect(TOKEN_IDENT).value.upper()
        if sense not in ['MAXIMIZE', 'MINIMIZE']:
            parser.error(PARSER_OBJ_SENSE_ERROR,
                         "Wrong name of objective sense.",
                         f"Dont specify sense. Must be MINIMIZE or MAXIMIZE but get: {sense}.")
        name = parser.expect(TOKEN_IDENT).value
        parser.expect(TOKEN_POINTPOINT)
        expr = parser.expr()
        parser.expect(TOKEN_SEMICOLON)
        return ObjectiveAST(pos, name, sense, expr)

    def skip_block(self, parser: BaseParser) -> None:
        parser.skip_to_semicolon_and_skip()



#  --------------------------------------------------------------------------------
#  Izok: A modeling system similar to GAMS for pyomo.
#  Copyright 2024,  Andrey Sergeevich Storozhenko, All rights reserved.
#  This software is distributed under the 3-clause BSD License.
#  --------------------------------------------------------------------------------

from typing import Union
from izok.extension.domain import Domain
from .position import Position


class AST:
    def __init__(self, pos: Position):
        self.pos = pos

    def to_est(self) -> Union['EST', None]:
        raise Exception("Must be override")


class EST:
    def __init__(self, ast: AST):
        self.ast = ast

    def __call__(self, model, domain: Domain):
        pass

    def __str__(self):
        return str(self.ast)



#  --------------------------------------------------------------------------------
#  Izok: A modeling system similar to GAMS for pyomo.
#  Copyright 2024,  Andrey Sergeevich Storozhenko, All rights reserved.
#  This software is distributed under the 3-clause BSD License.
#  --------------------------------------------------------------------------------

from .position import Position


class IzokException(Exception):
    def __init__(self, msg: str):
        super(IzokException, self).__init__(msg)


class Singleton:
    __instance = None

    def __init__(self):
        pass

    @classmethod
    def get_instance(cls):
        if not cls.__instance:
            cls.__instance = Singleton()
        return cls.__instance


class ExceptionConstructor(Singleton):
    def __init__(self):
        super(ExceptionConstructor, self).__init__()
        self.base_position = None
        self.scaner_source = ""

    def build_and_raise(self, current: Position, error_id: int, head_msg: str, tail_msg: str):
    
        if self.base_position is None:
            msg = "exception[E{:04d}]:{}".format(error_id, head_msg)
            msg += f"\n-->{self.scaner_source}:{current.line}:{current.pos}"
            num_count_in_line = len(f"{current.line}")
            line_number = f"\n{' ' * num_count_in_line}"
            msg += f"\n{current.line}|" + current.source
            msg += line_number + "|" + "-" * current.pos + "^"
            first = True
            for str_line in tail_msg.splitlines():
                if first:
                    msg += line_number + "=" + str_line
                    first = False
                else:
                    msg += line_number + " " + str_line
        else:
            msg = "exception[E{:04d}]:{}".format(error_id, head_msg)
            msg += f"\n-->{self.scaner_source}:{self.base_position.line}:{self.base_position.pos}"
            num_count_in_line = len(f"{current.line}")
            line_number = f"\n{' ' * num_count_in_line}"
            if self.base_position.line == current.line:
                msg += f"\n{self.base_position.line}|" + self.base_position.source
            else:
                msg += f"\n{self.base_position.line}|" + self.base_position.source
                if current.line - self.base_position.line > 1:
                    msg += line_number + "|" + '...'
                msg += f"\n{current.line}|" + self.base_position.source
            msg += line_number + "|" + "-" * current.pos + "^"
            first = True
            for str_line in tail_msg.splitlines():
                if first:
                    msg += line_number + "=" + str_line
                    first = False
                else:
                    msg += line_number + " " + str_line
    
        raise IzokException(msg)


def error(pos, error_id: int, head_msg: str, tail_msg: str):
    ExceptionConstructor().build_and_raise(pos, error_id, head_msg, tail_msg)


def set_base_position(value):
    ExceptionConstructor().base_position = value


def set_scaner_source(value):
    ExceptionConstructor().scaner_source = value



#  --------------------------------------------------------------------------------
#  Izok: A modeling system similar to GAMS for pyomo.
#  Copyright 2024,  Andrey Sergeevich Storozhenko, All rights reserved.
#  This software is distributed under the 3-clause BSD License.
#  --------------------------------------------------------------------------------

from sys import intern
from .position import Position

TOKEN_WHITE_SPACE = intern("whitespace")
TOKEN_NEW_LINE = intern("newline")
TOKEN_COMMENT = intern("comment")
TOKEN_IDENT = intern("ident")
TOKEN_LABEL = intern("label")
TOKEN_NUMBER = intern("number")

TOKEN_LPAREN = intern("lparen")
TOKEN_RPAREN = intern("rparen")
TOKEN_OPERATOR = intern("operator")
TOKEN_BAD_SYMBOL = intern("badsymbol")
TOKEN_END = intern("end")
TOKEN_SEMICOLON = intern("semicolon")
TOKEN_POINTPOINT = intern("pointpoint")
TOKEN_DOLLAR = intern("dollar")
TOKEN_EQNTYPE = intern("eqntype")
TOKEN_COMMA = intern("comma")


class Token:
    def __init__(self, pos: Position, type_str: str, value: str):
        self.pos = pos
        self.type = type_str
        self.value = value

    def __str__(self) -> str:
        if self.type == TOKEN_NEW_LINE or self.type == TOKEN_END:
            return f"[{self.pos.pos},{self.pos.line}] {self.type}"
        elif self.type == TOKEN_WHITE_SPACE:
            return f"[{self.pos.pos},{self.pos.line}] {self.type} symbol count {len(self.value)}"
        return f"[{self.pos.pos},{self.pos.line}] {self.type} {self.value}"



#  --------------------------------------------------------------------------------
#  Izok: A modeling system similar to GAMS for pyomo.
#  Copyright 2024,  Andrey Sergeevich Storozhenko, All rights reserved.
#  This software is distributed under the 3-clause BSD License.
#  --------------------------------------------------------------------------------

from typing import Tuple, Pattern, Callable, NamedTuple, List, Union
from .position import Position
from .token import Token


class ScanerRule(NamedTuple):
    pattern: Pattern[str]
    token: str
    action: Callable[[Position, str, str], Tuple[Token, Position]]


class BaseScaner:
    def __init__(self, text: str, rules: List[ScanerRule]):
        self.text = text
        self._text_by_line = (text+'\n').splitlines()
        self.pos = Position(self._text_by_line[0], 1, 0, 0)
        self._rules = rules

    def scan(self) -> Union[Token, None]:
        if self.pos.offset >= len(self.text):
            return None

        matches = []

        for r in self._rules:
            m = r.pattern.match(self.text, self.pos.offset)

            if m is None:
                continue

            value = m.group()

            if value == '':
                continue

            matches.append((r, m.group()))

        if len(matches) != 0:
            matches.sort(key=lambda x: -len(x[1]))
            rule, value = matches[0]
            token, self.pos = rule.action(self.pos, rule.token, value)
            return token

        token = Token(self.pos, "", self.text[self.pos.offset])
        self.pos.pos += 1
        self.pos.offset += 1
        return token




#  --------------------------------------------------------------------------------
#  Izok: A modeling system similar to GAMS for pyomo.
#  Copyright 2024,  Andrey Sergeevich Storozhenko, All rights reserved.
#  This software is distributed under the 3-clause BSD License.
#  --------------------------------------------------------------------------------

def build(model, text: str):
    from .parser import Parser
    parser = Parser(text)
    while True:
        ast = parser.parse()
        if ast is None:
            break

        est = ast.to_est()
        est.init(model)


__version__ = "0.0.0.dev0"



from setuptools import setup, find_packages

import json
import os

import codecs


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


def read_pipenv_dependencies(fname):
    """Get default dependencies from Pipfile.lock."""
    filepath = os.path.join(os.path.dirname(__file__), fname)
    with open(filepath) as lockfile:
        lockjson = json.load(lockfile)
        return [dependency for dependency in lockjson.get('default')]


if __name__ == "__main__":
    setup(
        name='Izok',
        version=os.getenv('PACKAGE_VERSION', get_version("./src/izok/__init__.py")),
        package_dir={'':'src'},
        packages=find_packages('src'),
        description='Izok is an open source Python-based software package that allows you to formulate an optimization problem in an easy-to-learn and efficient modeling language similar to other modeling languages AMPL, AIMMS and GAMS, and then automatically implement it in open source software package Pyomo. ',
    )


#  --------------------------------------------------------------------------------
#  Izok: A modeling system similar to GAMS for pyomo.
#  Copyright 2024,  Andrey Sergeevich Storozhenko, All rights reserved.
#  This software is distributed under the 3-clause BSD License.
#  --------------------------------------------------------------------------------

from typing import Union, List
from izok.extension.domain import Domain
from .errorid import *
from .error import error
from .extension.model import get_property_by_name, check_index_or_call
from .ast_est import AST, EST


_bin_op_lambda = {
    "+": lambda left, right: left + right,
    "-": lambda left, right: left - right,
    "/": lambda left, right: left / right,
    "*": lambda left, right: left * right,
    "**": lambda left, right: left ** right,
    "<>": lambda left, right: left != right,
    ">": lambda left, right: left > right,
    ">=": lambda left, right: left >= right,
    "<": lambda left, right: left < right,
    "<=": lambda left, right: left <= right,
    "=": lambda left, right: left == right,
    "=e=": lambda left, right: left == right,
    "=l=": lambda left, right: left <= right,
    "=g=": lambda left, right: left >= right,
    "and": lambda left, right: bool(left) and bool(right),
    "or": lambda left, right: bool(left) or bool(right),
    "xor": lambda left, right: bool(left) != bool(right),
}

_un_op_lambda = {
    "+": lambda l, r: r,
    "-": lambda l, r: - r,
    "not": lambda l, r: not bool(r),
}


class OperandEST(EST):
    def __init__(self, ast: AST, expr: EST, cond: Union[EST, None]):
        super(OperandEST, self).__init__(ast)
        self.expr = expr
        self.cond = cond

    def __call__(self, model, domain: Domain):
        if self.cond is not None:
            cond_value = self.cond(model, domain)
            try:
                cond_value = bool(cond_value)
            except Exception as err:
                error(
                    self.ast.pos,
                    GEN_TYPE_CONVERT_ERROR,
                    "Unable to reduce expression to boolean type.",
                    str(err)
                )
            if cond_value:
                return self.expr(model, domain)
            else:
                return 0
        else:
            return self.expr(model, domain)

#    def __str__(self):
#        if self.cond is None:
#            return f"{str(self.expr.ast)}"
#        else:
#            return f"{str(self.expr.ast)}${str(self.cond.ast)}"


class ExprEST(EST):
    def __init__(self, ast: AST, value: str, left: Union[EST, None], right: EST):
        super(ExprEST, self).__init__(ast)
        global _un_op_lambda, _bin_op_lambda
        self.ast = ast
        self.left = left
        self.right = right
        self.value = value
        if self.left is None:
            self.op = _un_op_lambda[value]
        else:
            self.op = _bin_op_lambda[value]

    def __call__(self, model, domain: Domain):
        left = None
        if self.left is not None:
            if isinstance(self.left, EST):
                left = self.left(model, domain)
            else:
                left = self.left
        if isinstance(self.right, EST):
            right = self.right(model, domain)
        else:
            right = self.right

        try:
            return self.op(left, right)
        except Exception as err:
            if self.left is None:
                error(self.right.ast.pos,
                      GEN_OPERATOR_ERROR,
                      f"Bad operand type for unary {self.value}: {type(right)}",
                      f"{err}")
            else:
                error(self.left.ast.pos,
                      GEN_OPERATOR_ERROR,
                      f"Unsupported operand type(s) for {self.value}: {type(left)} and {type(right)}",
                      f"{err}")

#    def __str__(self):
#        if self.left is None:
#            return f"[{self.value} {str(self.right.ast)}]"
#        else:
#            return f"[{str(self.right.ast)} {self.value} {str(self.left.ast)}]"


class IdentEST(EST):
    def __init__(self, ast: AST, ident: str):
        super(IdentEST, self).__init__(ast)
        self.ident = ident

#    def __str__(self):
#        return self.ident

    def __call__(self, model, domain: Domain):
        return get_property_by_name(model, domain, self.ast, self.ident)


class LabelEST(EST):
    def __init__(self, ast: AST, value: str):
        super(LabelEST, self).__init__(ast)
        self.value = value

#    def __str__(self):
#        return self.value

    def __call__(self, model, domain: Domain):
        return self.value


class NumberEST(EST):
    def __init__(self, ast: AST, value: str):
        super(NumberEST, self).__init__(ast)
        self.value = value

#    def __str__(self):
#        return self.value

    def __call__(self, model, domain: Domain):
        return float(self.value)


class ExprListEST(EST):
    def __init__(self, ast: AST, expr_list: List[EST]):
        super(ExprListEST, self).__init__(ast)
        self.expr_list = expr_list

#    def __str__(self):
#        if len(self.expr_list) == 1:
#            return f"{self.expr_list[0].ast}"
#        else:
#           return f"{','.join(str(e.ast) for e in self.expr_list)}"

    def __call__(self, model, domain: Domain):
        return [expr(model, domain) for expr in self.expr_list]


class SetListEST(EST):
    def __init__(self, ast: AST, idents: List[str]):
        super(SetListEST, self).__init__(ast)
        self.idents = idents

#    def __str__(self):
#        return str(self.ast)

    def __call__(self, model, domain: Domain):
        return [get_property_by_name(model, domain, self.ast, ident) for ident in self.idents]


class FunctionEST(EST):
    def __init__(self, ast: AST, ident: str, args: EST):
        super(FunctionEST, self).__init__(ast)
        self.ident = ident
        self.args = args

    def __call__(self, model, domain):
        prp = get_property_by_name(model, domain, self.ast, self.ident)
        args = self.args(model, domain)
        if check_index_or_call(prp):
            try:
                return prp.__getitem__(*args)
            except Exception as err:
                error(self.ast,
                      GEN_INDEXING_ERROR,
                      "Can't execute index operation",
                      str(err))
        else:
            try:
                return prp(*args)
            except Exception as err:
                error(self.ast,
                      GEN_INDEXING_ERROR,
                      "Can't call function",
                      str(err))
